#include "libckpt.h"
#include "copyright.h"

#define CHECK_RESULT(x)  if ( (x) == -1 ) {       \
                              _ckpt_flag = FALSE;    \
                              if ( _ckpt_fork_flag ) \
                                _exit(-1);           \
                              else                   \
                                return -1;           \
                            }

static off_t _stdin_seek_ptr;

/*
 *  function:     chld_handler()
 *  arguments:    none
 *  returns:      void
 *  side effects: Sets global variable _ckpt_wait_flag to FALSE.
 *                Disables future checkpoints by setting the global variable
 *                _ckpt_flag to FALSE if the child terminated abnormally
 *                or had a non-zero exit status.
 *  misc:         WIFEXITED is a macro defined in sys/wait.h returning TRUE
 *                if status was returned for a child that terminated normally.
 *                In this case we can call WEXITSTATUS(status), also defined
 *                in sys/wait.h, to fetch the low order 8 bits of the
 *                argument that the child passed to exit().
 */


#ifdef __linux
static void chld_handler(sig)
int sig;
#else
static void chld_handler()
#endif
{

  int status;

  signal(SIGCHLD, SIG_DFL);

  _ckpt_wait_flag = FALSE;

  wait ( &status );

  if ( ! WIFEXITED ( status ) || ( WEXITSTATUS ( status ) != 0 ) )
    _ckpt_flag = FALSE;

}

/*
 *  function:        checkpoint_here_()
 *  misc:            This is the function called from a FORTRAN program.
 *                   This function simply calls checkpoint_here().
 */

int checkpoint_here_()
{

  return checkpoint_here();

}

/*
 *  function:        checkpoint_here()
 *  arguments:       none
 *  returns:         0 if returning from taking a checkpoint, 1 if returning
 *                   from a recovery, -1 on error.
 *  local variables: stop_addr  :  sbrk(0), used to see if break has moved.
 *                   temp_break :  used to call include if break has moved.
 *                   rflag      :  return value.
 *  side effects:    fills global setjmp buffer env by calling setjmp()
 *                   adds info to global included struct if break has moved
 *                   sets the seek pointer in open file table by calling gseek
 *                   writes a new checkpoint file
 *                   sync's the file system
 *                   increments the global variable _ckpt_filenum
 *                   sets the global variable _old_break to sbrk(0)
 *                   calls protect_all(), which has global side effects
 *                   opens files in file table if recovering
 */

int checkpoint_here()
{

  caddr_t stop_addr;                 /* stoping address of heap */
  char *temp_break;                  /* temp variable for calling include() */
  int rflag = FALSE;                 /* return value */

  _ckpt_num++;

  if ( check_flags() == -1 )
    return -1;

  /* cancel timeout */
  alarm(0);

  if ( _ckpt_verbose_flag ) {
    fprintf ( stderr, "CKP %d %d: beginning\n", _ckpt_num, time(0));
  }

  /* set seek pointers in file table for use during recovery */

  gseek();

  /* if the break has moved, record current break */

  stop_addr = (caddr_t) sbrk(0);
  if ( _ckpt_i_old_break < stop_addr ) {
    temp_break = _ckpt_i_old_break;
    _ckpt_i_old_break = stop_addr;
    include_bytes ( temp_break, stop_addr - temp_break );
  }

  /* if I'm about to fork(), initialize some stuff */

  if ( _ckpt_fork_flag ) {
    _ckpt_wait_flag = TRUE;
    signal(SIGCHLD, chld_handler);
    fflush(stdout);
    fflush(stderr);
  }

  /* take the checkpoint */

  if ( ( _ckpt_fork_flag && ! fork() ) || ! _ckpt_fork_flag )
    rflag = take_ckpt();

  CHECK_RESULT(cleanup(rflag));

  _ckpt_time = time(0);

  return rflag;

}

/*
 *  function:        check_flags()
 *  arguments:       none
 *  returns:         0 on success, -1 on failure and sets errno.
 *  local variables: now -- the time, used for verbose mode.
 *                   retval -- holds the return value.
 *                   rtime -- time remaining until next asynchronous checkpoint
 *  side effects:    sets errno if there is an error.
 *  calls:           nothing
 *  called from:     checkpoint_here()
 */

static int check_flags()
{
  int now = time(0);
  int retval = 0;
  int rtime;

  if ( ! _ckpt_flag ) {
    errno = ENOCKPT;
    if ( _ckpt_verbose_flag )
      fprintf ( stderr, "CKP %d : checkpointing currently disabled\n", now);
    retval = -1;
  }

  if ( ! retval && _ckpt_wait_flag ) {
    errno = ECHILD;
    if ( _ckpt_verbose_flag ) {
      fprintf ( stderr, "CKP %d : previous main memory checkpoint ", now ) ;
      fprintf ( stderr, "still in progress\n" );
    }
    retval = -1;
  }

  if ( ! retval && ! _ckpt_enable ) {
    errno = ETOOSOON;
    rtime = _ckpt_min_time - (now - _ckpt_time);
    if ( _ckpt_verbose_flag ) {
      fprintf ( stderr, "CKP %d : too soon between checkpoints; ", now );
      fprintf ( stderr, "%d seconds remaining\n", rtime );
    }
    retval = -1;
  }

  return retval;

}

/*
 *  function:        gseek()
 *  arguments:       none
 *  returns:         void
 *  local variables: fd -- index into _ckpt_ftable[]
 *                   fptr -- pointer to _ckpt_ftable[fd]
 *  side effects:    sets seek_ptr field in _ckpt_ftable for each open file
 *                   sets global _stdin_seek_ptr.
 *  calls:           nothing
 *  called from:     checkpoint_here()
 */

static void gseek()
{
  int fd;
  ftable_entry *fptr;

  for (fd = 0; fd < _CKPT_FTABLE_SIZE; fd++) {
    fptr = &_ckpt_ftable[fd];
    if (fptr->inuse)
	fptr -> seek_ptr = lseek(fptr -> fd, 0, SEEK_CUR);
  }

  _stdin_seek_ptr = lseek(0, 0, SEEK_CUR);
}

/*
 *  function:        take_ckpt()
 *  arguments:       none
 *  returns:         0 on success, -1 on failure
 *  local variables: now -- the time, used for verbose mode.
 *                   nchunks -- the number of chunks in the current ckpt file.
 *                   filename -- the name of the current ckpt file (e.g. ckpt.0)
 *                   tempname -- the name of a temp file.
 *                   rflag -- holds return value.
 *                   fd -- file descriptor for ckpt file
 *                   result -- checks return value of called functions.
 *  side effects:    takes a checkpoint, possibly coalescing ckpt files, or
 *                   opens files if recovering.
 *  calls:           write_header()
 *                   write_i_heap()
 *                   write_heap()
 *                   write_stack()
 *                   gsync()
 *                   _ckpt_coalesce_files()
 *                   open_files()
 *  called from:     checkpoint_here()
 */

static int take_ckpt()
{

  int now = time(0);
  long nchunks = 0;
  char filename[100];
  char tempname[100];
  int rflag = FALSE;
  int fd;
  int result;

  if ( ! setjmp(_ckpt_env) ) {

    sprintf ( tempname, "%s/%s.%s", _ckpt_dir, _ckpt_filename, "temp" );

    remove ( tempname );

    fd = OPEN_M ( tempname, O_WRONLY | O_CREAT, 0644 );

    CHECK_RESULT(fd)

    CHECK_RESULT(write_header(fd))

    if ( _ckpt_inc_flag )
      result = i_write_heap(fd);
    else
      result = write_heap(fd);

    CHECK_RESULT(result)

    nchunks += result;

    result = write_stack(fd);
    CHECK_RESULT(result)
    nchunks += result;

    lseek(fd, POINTER_SIZE * 2, SEEK_SET); 
    CHECK_RESULT(write(fd, (char *)&nchunks, POINTER_SIZE))

    CLOSE(fd);
    
    sprintf ( filename, "%s/%s.%d", _ckpt_dir, _ckpt_filename, _ckpt_filenum );

    rename ( tempname, filename );

    CHECK_RESULT(gsync())

    if ( _ckpt_coa_flag && _ckpt_max_files && _ckpt_max_files == _ckpt_filenum ) {
      _ckpt_coalesce_files(_ckpt_dir);
      _ckpt_filenum = 0;
    }

    if ( _ckpt_fork_flag ) {
      if ( _ckpt_verbose_flag ) {
        fprintf ( stderr, "CKP %d : checkpoint complete, child dying\n", now);
      }
      _exit(0);
    }

  }

  else {
    open_files();
    rflag = TRUE;
    if ( _ckpt_fork_flag )
      _ckpt_wait_flag = FALSE;
    signal(SIGALRM, _ckpt_alrm_handler);
  }

  return rflag;

}

/*
 *  function:        write_header()
 *  arguments:       fd          -- file descriptor for open ckpt file
 *  returns:         0 on success, -1 on error and sets _ckpt_flag to FALSE
 *  local variables: dummy_first -- top variable on stack on DEC machines
 *                   top_stack   -- actual address of the top of the stack
 *                   retval      -- holds return value.
 *                   nchunks     -- place holder, filled in later
 *                   dummy_last  -- top of stack on SUN machines
 *  side effects:    writes useful info to ckpt file
 *  calls:           write()
 *  called from:     take_ckpt()
 *  misc:            the address of the local variable dummy_first or
 *                   dummy_last is refered to by TOP_STACK, defined
 *                   in machdep.h
 */

static int write_header(fd)
int fd;
{

  int dummy_first;
  unsigned long top_stack, stop_addr;
  int retval = 0;
  int nchunks = 0;
  int dummy_last;

  top_stack = TOP_STACK;
  stop_addr = (unsigned long)sbrk(0);

  if ( ! retval && write(fd, (char *)&top_stack, POINTER_SIZE) != POINTER_SIZE )
    retval = -1;

  if ( ! retval && write(fd, (char *)&stop_addr, POINTER_SIZE) != POINTER_SIZE )
    retval = -1;

  if ( ! retval && write(fd, (char *)&nchunks, POINTER_SIZE) != POINTER_SIZE)
    retval = -1;

  if ( retval )
    _ckpt_flag = FALSE;

  return retval;

}

/*
 *  function:        i_write_heap()
 *  arguments:       fd       -- a file descriptor for open ckpt file.
 *  returns:         The number of chunks written on success, or -1 on failure
 *  local variables: list_ptr -- pointer to traverse _ckpt_inc list
 *                   end_ptr  -- pointer to the "end" of the list
 *                   addr     -- address of page being written to disk
 *                   size     -- size of page being written to disk
 *                   nchunks  -- holds return value
 *                   pn       -- the page number
 *  side effects:    for each "included" chunk of memory in the heap, writes
 *                   all dirty pages to the open ckpt file
 *  calls:           isnew()
 *                   isdirty()
 *                   write_chunk()
 *  called from:     take_ckpt()
 */

static int i_write_heap(fd)
int fd;
{

  Dlist list_ptr, end_ptr;
  char *addr;
  char *stop;
  long size;
  int nchunks = 0;
  int pn;

  /* for each "included" chunk of memory */

  
  end_ptr = _ckpt_inc->main->blink;
  for(list_ptr = _ckpt_inc->main->flink->flink; list_ptr != end_ptr; 
	list_ptr = list_ptr -> flink) {

    if ( nchunks == -1 )
      break;

    stop = list_ptr -> stop;

    /*
     *  if the first (possibly partial) page of the
     *  current "included" chunk is dirty, write it
     *
     *  The calculation for size is complex because this first page
     *  may NOT extend all the way to the next page boundary.  If it
     *  does, it gets the first argument to the min macro, otherwise
     *  it gets the second, "stop - addr".
     */

    pn = (long)((long)(list_ptr -> addr - START_ADDR) / GPAGESIZE);
    if ( isnew(pn) || isdirty(pn) ) {
      addr = list_ptr -> addr;
      size = min ( ((((long)((long)addr / GPAGESIZE) + 1) * GPAGESIZE) -
                    (long)addr),
                   (stop - addr) );
      if ( write_chunk(addr, size, fd) == -1 ) {
        nchunks = -1;
      } else {
        nchunks++;
      }
    }

    /* for each whole page in the current "included" chunk */


    if ( nchunks != -1 )
      for (pn++; ((pn + 1) * GPAGESIZE) + START_ADDR < (long)stop; pn++)
        if ( isnew(pn) || isdirty(pn) ) {
          addr = (char *) (pn * GPAGESIZE + START_ADDR);
          if ( write_chunk(addr, GPAGESIZE, fd) == -1) {
            nchunks = -1;
            break;
          }
          else {
            nchunks++;
          }
        }

    /*
     *  if the last (possibly partial) page is dirty, write it
     *  find addr and size first, because there may not be a last page
     */

    addr = (char *)(pn * GPAGESIZE + START_ADDR);
    size = stop - addr;
    if ( addr <= stop && ( isnew(pn) || isdirty(pn) ) ) {
      if ( write_chunk(addr, size, fd) == -1 )
        nchunks = -1;
      else
        nchunks++;
    }

  }

  return nchunks;

}

/*
 *  function:        write_heap()
 *  arguments:       fd -- a file descriptor for the open ckpt file
 *  returns:         the number of chunks written on success, -1 on failure.
 *  local variables: inc_ptr -- pointer to traverse _ckpt_inc list
 *                   end_ptr -- pointer to "end" of _ckpt_inc list
 *                   nchunks -- holds return value
 *  side effects:    writes all "included" chunks of the heap to the ckpt file
 *  calls:           write_chunk()
 *  called from:     take_ckpt()
 */

static int write_heap(fd)
int fd;
{

  Dlist inc_ptr, end_ptr;            /* ptr to traverse include/exclude list */
  int nchunks = 0;                   /* return value; nchunks if ok, else -1 */

  end_ptr = _ckpt_inc->main->blink;
  for ( inc_ptr = _ckpt_inc->main->flink->flink; inc_ptr != end_ptr; 
	inc_ptr = inc_ptr -> flink ) {
    if ( write_chunk ( inc_ptr -> addr, inc_ptr -> size, fd ) == -1 ) {
      nchunks = -1;
      break;
    }
    nchunks++;
  }

  return nchunks;

}

/*
 *  function:        write_stack()
 *  arguments:       fd -- a file descriptor for the open ckpt file.
 *  returns:         The number of chunks written on success, -1 on failure.
 *  local variables: dummy_first -- top variable on the stack on DEC machines.
 *                   nchunks     -- holds return value.
 *                   dummy_last  -- top variabel on the stack on SUN machines.
 *  side effects:    writes the stack to the open ckpt file
 *  calls:           write_chunk()
 *  called from:     take_ckpt()
 *  misc:            on SPARCs, there is an area above the stack which holds
 *                   open file info.
 *                   the address of either the local variable dummy_first
 *                   or dummy_last is refered to in TOP_STACK, defined in
 *                   machdep.h
 */

static int write_stack(fd)
int fd;
{

  int dummy_first;
  int nchunks = 0;
  int dummy_last;


#ifdef	sparc
  if ( _ckpt_lang_flag == 0 ) {
    if ( write_chunk(FILE_AREA, FILE_AREA_SIZE, fd) == -1 )
      nchunks = -1;
    else
      nchunks++;
  }
#endif

  if ( nchunks != -1 ) {
#ifdef STACK_GROWS_UP
    if ( write_chunk((char *)TOP_STACK, BOTTOM_STACK - TOP_STACK, fd) == -1 )
#else
    if ( write_chunk((char *)BOTTOM_STACK, TOP_STACK - BOTTOM_STACK, fd) == -1 )
#endif
      nchunks = -1;
    else
      nchunks++;
  }

  return nchunks;

}

/*
 *  function:        gsync()
 *  arguments:       none
 *  returns:         0 on success, -1 on failure
 *  local variables: fd -- index into the file table
 *                   fptr -- pointer to _ckpt_ftable[fd]
 *  side effects:    sync's the file system
 *  calls:           fsync()
 *  called from:     take_ckpt()
 */

static int gsync()
{
  int fd;
  ftable_entry *fptr;

  for (fd = 0; fd < _CKPT_FTABLE_SIZE; fd++)
    if (_ckpt_ftable[fd].inuse && fsync(fd) == -1 ) return -1;

  return 0;

}

/*
 *  function:        write_chunk()
 *  arguments:       addr -- the beginning address of the chunk to write
 *                   size -- the size of the chunk to write
 *                   fd   -- the file descriptor for the open ckpt file
 *  returns:         0 on success, -1 on failure.
 *  local variables: none
 *  side effects:    writes a chunk to the ckpt file
 *  calls:           write()
 *  called from:     take_ckpt()
 *                   i_write_heap()
 *                   write_heap()
 *                   write_stack()
 */

static int write_chunk(addr, size, fd)
char *addr;
long size;
int fd;
{

  if ( write(fd, (char *)&addr, POINTER_SIZE) != POINTER_SIZE ) {
    _ckpt_flag = FALSE;
    return -1;
  }

  if ( write(fd, (char *)&size, POINTER_SIZE) != POINTER_SIZE ) {
    _ckpt_flag = FALSE;
    return -1;
  }

  if ( write(fd, addr, size) != size ) {
    _ckpt_flag = FALSE;
    return -1;
  }

  return 0;

}

/*
 *  function:        open_files()
 *  arguments:       none
 *  returns:         void
 *  local variables: ptr -- a pointer into the file table, _ckpt_ftable[i]
 *                   dptr -- a pointer to the dup list (handles dups)
 *                   i -- loop counter, indexes _ckpt_ftable
 *                   fd -- checks return val of syscall, ensures file opened ok
 *  side effects:    opens files from the global _ckpt_ftable, handles dups,
 *                   lseeks to the correct point in the file.
 *  calls:           syscall()
 *                   lseek()
 *  called from:     take_ckpt()
 */

static void open_files()
{

/*
 *  The attribute field of the file table entry must have the _TRUNC
 *  flag removed.  This flag is 0x0400, so using a bitwise and with
 *  0xfffffbff removes it if it is set.
 */

  ftable_entry *ptr, *dptr;
  int i, fd;

  for (i = 0 ; i < _CKPT_FTABLE_SIZE; i++) {

    ptr = &_ckpt_ftable[i];
    if (ptr->inuse) {
      
      dptr = ptr->dup_list;
      while(dptr->fd > ptr->fd) dptr = dptr->dup_list;

      if (dptr == ptr) {

      	fd = OPEN_M(ptr -> filename, ptr -> attribute & 0xfffffbff, ptr -> mode);

        if ( fd == -1 ) {
          fprintf(stderr, " cannot re-open file %s during recovery\n",
          		    ptr -> filename);
          exit(-1);
        }
        if (fd != ptr->fd) {
	  DUP2(fd, ptr->fd );
	  CLOSE(fd);
        }

      } else

      	DUP2(dptr->fd, ptr->fd );

      lseek(ptr -> fd, ptr -> seek_ptr, SEEK_SET);
    }
  }

}

/*
 *  function:        protect_all()
 *  arguments:       none
 *  returns:         0 on success, -1 on failure
 *  local variables: lb    -- lower bound of the page where the pagelist resides
 *                   ub    -- upper bound of the page where the pagelist resides
 *                   addr  -- used as an argument to _ckpt_markpage() when
 *                            the page holding the pagelist is set to dirty
 *                   start -- Beginning address of area to set READ ONLY.
 *                            Does not include first page, since this page
 *                            cannot be mprotect()'d on SparcStations.
 *                   stop  -- Value of sbrk(0).
 *                   i     -- loop index.
 *  side effects:    sets up incremental checkpointing by malloc()'ing a
 *                   table to keep track of dirty pages, calling mprotect()
 *                   set all memory aside from the pagetable to READ ONLY,
 *                   and calling signal to catch the resulting SIGSEGV signals.
 *  calls:           mprotect()
 *                   free()
 *                   malloc()
 *                   fprintf()
 *                   signal
 *                   sbrk()
 *                   _ckpt_markpage()
 *  called from:     cleanup()
 */

static int protect_all()
{

  unsigned long lb, ub;
  unsigned long addr;
  unsigned long start, stop;
  int i;

  start = START_ADDR + GPAGESIZE;
  _ckpt_d_old_break = (char *) sbrk(0);
  stop = (long)_ckpt_d_old_break;

  if ( stop % GPAGESIZE )
    stop += stop % GPAGESIZE;

  mprotect((char *)start, stop - start, PROT_READ | PROT_WRITE);

  free(_ckpt_pagelist);

  _ckpt_np = (long)((stop - START_ADDR) / GPAGESIZE);
  if (stop % GPAGESIZE) _ckpt_np++;

  _ckpt_pagelist = (unsigned char *)malloc(_ckpt_np);

  if ( _ckpt_pagelist == NULL ) {
    fprintf(stderr, "out of memory\n");
    _ckpt_flag = FALSE;
    errno = ENOMEM;
    return (-1);
  }

  for ( i = 0; i < _ckpt_np; i++ ) _ckpt_pagelist[i] = 0;

  lb = (unsigned long)((unsigned long)_ckpt_pagelist / GPAGESIZE) * GPAGESIZE;
  ub = ((unsigned long)(((unsigned long)_ckpt_pagelist + _ckpt_np) / GPAGESIZE) + 1) * GPAGESIZE;

#ifndef __linux
  signal(SIGSEGV, _ckpt_pgfault_handler);
  signal(SIGBUS, _ckpt_pgfault_handler);
#endif

  stop = (long)sbrk(0);

  if ( stop % GPAGESIZE )
    stop += GPAGESIZE - (stop % GPAGESIZE);

  mprotect((char *)start, stop - start, PROT_READ);
  mprotect((char *)lb, ub - lb, PROT_READ | PROT_WRITE);

  _ckpt_markpage((unsigned long)START_ADDR);

  addr = (unsigned long)_ckpt_pagelist;
  if ((addr % GPAGESIZE) != 0) _ckpt_markpage(addr);
  if ((addr + _ckpt_np * sizeof(unsigned char)) % GPAGESIZE != 0)
            _ckpt_markpage(addr + _ckpt_np * sizeof(unsigned char));

  return 0;

}

/*
 *  function:        cleanup()
 *  arguments:       rflag - the return value so far
 *  returns:         0 on success, -1 on error.
 *  local variables: now -- the time, used for verbose mode
 *                   result -- holds return value
 *                   sav_ptr -- pointer for traversing _ckpt_sav list
 *                   end_ptr -- pointer to "end" of _ckpt_sav list
 *  side effects:    sets up incremental checkpointing
 *                   "excludes" memory on "pending" list
 *                   updates _ckpt_filenum
 *                   resets the alarm time
 *  calls:           protect_all()
 *                   exclude_bytes()
 *                   ml_free_list()
 *                   make_ml()
 *                   alarm()
 *                   fprintf()
 *  called from:     checkpoint_here()
 */

static int cleanup(rflag)
int rflag;
{
  int now = time(0);
  int result = 0;
  Dlist sav_ptr, end_ptr;

  /*
   *  if using incremental mode, mark all pages clean, reset
   *  page protection to READ_ONLY
   */

  if ( _ckpt_inc_flag )
    result = protect_all();

  sav_ptr = _ckpt_sav->main->flink->flink;
  end_ptr = _ckpt_sav->main->blink;

  while ( sav_ptr != end_ptr ) {
    exclude_bytes ( sav_ptr -> addr, sav_ptr -> size , TRUE);
    sav_ptr = sav_ptr -> flink;
  }
  ml_free_list(_ckpt_sav);
  _ckpt_sav = make_ml();

  if ( ! _ckpt_coa_flag )
    _ckpt_filenum = 0;
  else if ( _ckpt_max_files && _ckpt_max_files == _ckpt_filenum )
    _ckpt_filenum = 1;
  else
    _ckpt_filenum++;

  /* reset alarmtime */
  if ( _ckpt_min_time ) {
    _ckpt_enable = FALSE;
    alarm(_ckpt_min_time);
  }
  else {
    _ckpt_enable = TRUE;
    if ( _ckpt_max_time )
      alarm(_ckpt_max_time);
  }

  if ( _ckpt_verbose_flag )
    if ( rflag == 1 )
      fprintf(stderr, "CKP %d : recovery complete, resuming execution\n", now);
    else if ( _ckpt_fork_flag )
      fprintf(stderr, "CKP %d : parent exiting, resuming execution\n", now);
    else
      fprintf(stderr, "CKP %d : checkpoint complete, resuming execution\n", now);

  return result;

}

/*
 *  function:     isnew()
 *  arguments:    pn - page number; the first page of the heap is 0
 *  returns:      TRUE if any portion of page pn lies past the old break,
 *                FALSE otherwise
 *  side effects: none known
 *  calls:        nothing
 *  called from:  i_write_heap()
 */

static int isnew(pn)
int pn;
{

  if ( (pn + 1) * GPAGESIZE + START_ADDR > (long)_ckpt_d_old_break )
    return TRUE;
  else
    return FALSE;

}

/*
 *  function:        isdirty()
 *  arguments:       pn -- a page number
 *  returns:         1 if the page is dirty, 0 otherwise
 *  local variables: none
 *  side effects:    none known
 *  calls:           nothing
 *  called from:     i_write_heap()
 */

static int isdirty(pn)
int pn;
{

  return (int) (_ckpt_pagelist[pn]);

}
