#include "libckpt.h"
#include "copyright.h"

#define	isempty(x)	((x) == NULL)

/*
 *  function:        include_bytes_()
 *  arguments:       addr -- the beginning address of the chunk to include
 *                   sizep -- a pointer to the size of the chunk (FORTRAN)
 *  returns:         0 on success, -1 on failure
 *  local variables: none
 *  side effects:    Adds the indicated chunk to _ckpt_inc.
 *  calls:           include_bytes()
 *  called from:     A FORTRAN program
 */

int include_bytes_(addr, sizep)
char *addr;
unsigned long *sizep;
{

  return include_bytes(addr, *sizep);

}

/*
 *  function:        exclude_bytes_()
 *  arguments:       addr -- the beginning address of the chunk to include
 *                   sizep -- a pointer to the size of the chunk (FORTRAN)
 *                   start_nowp -- a pointer to the start_now parameter
 *  returns:         0 on success, -1 on failure
 *  local variables: none
 *  side effects:    Removes the indicated chunk from _ckpt_inc.
 *  calls:           exclude_bytes()
 *  called from:     A FORTRAN program
 */

int exclude_bytes_(addr, sizep, start_nowp)
char *addr;
unsigned long *sizep;
int *start_nowp;
{
  return exclude_bytes(addr, *sizep, *start_nowp);
}

/*
 *  function:        exclude_bytes()
 *  arguments:       addr      -- The address of the chunk to exclude
 *                   size      -- The size of the chunk to exclude
 *                   start_now -- If TRUE, chunk is removed from  _ckpt_inc,
 *                                if FALSE, chunk is added to _ckpt_sav.
 *  returns:         0 on success, -1 on failure.
 *  local variables: result -- holds return value
 *  side effects:    adds to _ckpt_sav or removes from _ckpt_inc depending
 *                   on the value of the start_now parameter.
 *  calls:           exclude()
 *                   include()
 *  called from:     exclude_bytes_()
 *                   user code
 */

int exclude_bytes(addr, size, start_now)
char *addr;
unsigned long size;
int start_now;
{

  int result;

  if (!_ckpt_exclude_flag) return;

  CKPT_MUTEX_GET

  if ( start_now )
    result = exclude(_ckpt_inc, addr, size);
  else
    result = include(_ckpt_sav, addr, size);

  CKPT_MUTEX_REL

  return result;

}


/*
 *  function:        include_bytes()
 *  arguments:       addr -- The address of the chunk to include
 *                   size -- The size of the chunk to include
 *  returns:         0 on success, -1 on failure
 *  local variables: result -- holds return value
 *  side effects:    adds the indicated chunk to _ckpt_inc
 *  calls:           include()
 *  called from:     include_bytes_()
 *                   user's code
 */

int include_bytes(addr, size)
char *addr;
unsigned long size;
{

  int result;

  if ( ! _ckpt_flag ) {
    errno = ENOCKPT;
    return - 1;
  }

  if ( (long)addr < (long)START_ADDR || (long)addr + size > (long)sbrk(0) ) {
    errno = EFAULT;
    return -1;
  }

  CKPT_MUTEX_GET
  result = include(_ckpt_inc, addr, size);
  CKPT_MUTEX_REL

  return result;

}


/*
 *  function:        include()
 *  arguments:       list -- the list to add to (_ckpt_inc or _ckpt_sav)
 *                   addr -- the address of the chunk to include
 *                   size -- the size of the chunk to include
 *  returns:         0 on success, -1 on flure
 *  local variables: list_ptr -- a pointer into the list
 *                   stop -- the address of the end of the chunk
 *  side effects:    adds the indicated chunk to the list
 *  calls:           ml_find()
 *                   coaalesce()
 *                   ml_add()
 *  called from:     include_bytes()
 *                   exclude_bytess()
 */

static int include(list, addr, size)
Mlist list;
char *addr;
unsigned long size;
{
  Dlist list_ptr;
  char *stop;

  stop = (char *)((long)addr + size);

  list_ptr = ml_find(list, addr);

  if ( list_ptr -> stop >= addr ) { /* cases 2.1, 2.2 */
     list_ptr -> stop = max(list_ptr -> stop, stop);
     list_ptr -> size = list_ptr -> stop - list_ptr -> addr;
     coalesce( list, list_ptr );
  } else { /* case 2.3 */
     ml_add(list, list_ptr, addr, size);
     list_ptr = list_ptr -> flink;
     coalesce( list, list_ptr);
  }
  return 0;
}


/*
 *  function:        coalesce()
 *  arguments:       list -- one of _ckpt_save or _ckpt_inc
 *                   list_ptr -- a pointer into the list
 *  returns:         void
 *  local variables: none
 *  side effects:    "coalesces" the list pointed to by list,
 *                   beginning at list_ptr
 *  calls:           ml_delete()
 *  called from:     include()
 */

static void coalesce(list, list_ptr)
Mlist list;
Dlist list_ptr;
{

  while ( list_ptr -> flink -> stop <= list_ptr -> stop) {
    ml_delete(list, list_ptr -> flink);
  }

  if ( list_ptr -> flink -> addr <= list_ptr -> stop ) {
    list_ptr -> stop = list_ptr -> flink -> stop;
    list_ptr -> size = list_ptr -> stop - list_ptr -> addr;
    ml_delete(list, list_ptr -> flink);
  }

}

/*
 *  function:        exclude()
 *  arguments:       list -- the list (_ckpt_inc or _ckpt_sav) to modify
 *                   addr -- the address of the chunk to exclude
 *                   size -- the size of the chunk to exclude
 *  returns:         0 on success, -1 on failure
 *  local variables: list_ptr  -- a pointer into the list
 *                   stop      -- the address of the "end" of the chunk
 *                   temp_addr -- used to call include() if the break has moved.
 *  side effects:    "excludes" a chunk of memory by possibly removing info
 *                   from the list passed in as a first argument
 *  calls:           include()
 *                   ml_find()
 *                   ml_delete()
 *                   ml_add()
 *                   exclude()
 *  called from:     exclude() -- recursive
 *                   exclude_bytes()
 */

static int exclude(list, addr, size)
Mlist list;
char *addr;
unsigned long size;
{

  Dlist list_ptr;
  char *stop;
  char *temp_addr;

  if ( ! _ckpt_flag ) {
    errno = ENOCKPT;
    return -1;
  }

  if ( addr + size > _ckpt_i_old_break) {
    temp_addr = _ckpt_i_old_break;
    _ckpt_i_old_break = (char *) sbrk(0);
    include( list, temp_addr, _ckpt_i_old_break - temp_addr );
  }

  if ( (long)addr < (long)START_ADDR || addr + size > _ckpt_i_old_break ) {
    errno = EFAULT;
    return -1;
  }

  stop = (char *)((long)addr + size);

  list_ptr = ml_find(list, addr);

  if ( addr == list_ptr -> addr ) { /* case 2.1 */
    if ( stop < list_ptr -> stop ) { /* case 2.1.1 */
      list_ptr -> addr = stop;
      list_ptr -> size = list_ptr -> stop - list_ptr -> addr;
    } else if ( stop == list_ptr -> stop )  { /* case 2.1.2 */
      ml_delete(list, list_ptr);
    } else { /* case 2.1.3 */
      ml_delete(list, list_ptr);
      exclude(list, addr, size);
    }
  } else if ( addr < list_ptr -> stop) { /* case 2.2 */
    if ( stop < list_ptr -> stop ) { /* case 2.2.1 */
      ml_add(list, list_ptr, stop, list_ptr -> stop - stop);
      list_ptr -> stop = addr;
      list_ptr -> size = list_ptr -> stop - list_ptr -> addr;
    }
    else if ( stop == list_ptr -> stop ) { /* case 2.2.2 */
      list_ptr -> stop = addr;
      list_ptr -> size = list_ptr -> stop - list_ptr -> addr;
    }
    else { /* case 2.2.3 */
      list_ptr -> stop = addr;
      list_ptr -> size = list_ptr -> stop - list_ptr -> addr;
      exclude(list, addr, size);
    }
  }
  else { /* case 2.3 */
    addr = list_ptr -> flink -> addr;
    if ( addr < stop ) {
      exclude(list, addr, stop - addr);
    }
  }

  return 0;

}
