#include "libckpt.h"
#include "copyright.h"

/*
 *  FUNCTION CALL TREE
 *
 *  ckpt_setup_()
 *        _ckpt_setup()
 *              set_flags()
 *                    check_for_recovery()
 *                          _ckpt_get_path()
 *                                read_dotfile()
 *                                      init_flag_array()
 *                                      convert()
 *                                            isnum()
 *                          _ckpt_recover()
 *                                SEE recover.c
 *              set_defaults()
 *                    read_dotfile()
 *                          init_flag_array()
 *                          convert()
 *                                isnum()
 *              print_defaults()
 *              set_globals()
 *                    make_ml()
 *              check_files()
 *              include_bytes()
 *                    SEE include.c
 *              set_incremental()
 *                    _ckpt_markpage()
 *              set_alarm()
 */

/*
 *  function:        _ckpt_pgfault_handler()
 *  arguments:       sig -- the signal number
 *                   code -- an unused parameter
 *                   scp -- pointer to sigcontext structure defined in signal.h
 *                   addr -- additional address information
 *  returns:         void
 *  local variables: none
 *  side effects:    sets the offending page to READ / WRITE
 *                   marks the offending page as dirty
 *  calls:           _ckpt_markpage()
 *  called from:     potentially, anywhere
 */

#ifndef __linux
void _ckpt_pgfault_handler(sig, code, scp, addr)
long sig, code;
struct sigcontext *scp;
char *addr;
{

  mprotect((char *)((BADADDR / GPAGESIZE) * GPAGESIZE), GPAGESIZE, PROT_READ | PROT_WRITE);

  _ckpt_markpage(BADADDR);

#ifdef _SGI_SOURCE
  signal(SIGSEGV, _ckpt_pgfault_handler);
#endif

}
#endif

/*
 *  function:        _ckpt_alrm_handler()
 *  arguments:       none
 *  returns:         void
 *  local variables: none
 *  side effects:    enables checkpointing if the min time has expired
 *                   takes a checkpoint is the max time has expired
 *                   writes to stderr if verbose mode is on
 *  calls:           checkpoint_here() if max time has expired
 *  called from:     potentially, anywhere
 */

#ifdef __linux
void _ckpt_alrm_handler(sig)
int sig;
#else
void _ckpt_alrm_handler()
#endif
{

  alarm(0);
  if ( _ckpt_enable == FALSE ) {  /*  min time expired, enable checkpointing  */

    if ( _ckpt_verbose_flag ) {
      fprintf ( stderr, "CKP %d : %d seconds has passed since last checkpoint;",
                time(0), _ckpt_min_time );
      fprintf ( stderr, " enabling checkpointing\n");
      if ( _ckpt_max_time ) {
        fprintf ( stderr, "%d seconds remaining until interupt ",
                  _ckpt_max_time - _ckpt_min_time );
        fprintf ( stderr, "based checkpoint\n");
      }
    }

    _ckpt_enable = TRUE;
    if ( _ckpt_max_time )
      alarm(_ckpt_max_time - _ckpt_min_time);

  }

  else {  /*  max time expired, checkpoint; timer reset inside checkpoint()  */

    if ( _ckpt_verbose_flag ) {
      fprintf ( stderr, "CKP %d : %d seconds has passed since last checkpoint;",
                time(0), _ckpt_max_time );
      fprintf ( stderr, "taking interupt based checkpoint\n");
    }

    if ( _ckpt_mutex != CKPT_ALLOW ) { /* checkpoint disallowed, delay */
	_ckpt_mutex = CKPT_DELAY;	
	return;
    }
    checkpoint_here();

  }

}

/*
 *  function:        print_defaults()
 *  arguments:       none
 *  returns:         void
 *  local variables: bool, an array of char *'s with 2 elements, "off" and "on"
 *  side effects:    prints some stuff to stderr if verbose mode is on
 *  calls:           nothing
 *  called from:     _ckpt_setup()
 */

static void print_defaults()
{

  char *bool[2];
  bool[0] = "off";
  bool[1] = "on";

  if (_ckpt_verbose_flag) {
    fprintf(stderr, "BEGIN LIBCKPT DEFAULTS\n");
    fprintf(stderr, "    checkpointing %s\n", bool[_ckpt_flag]);
    fprintf(stderr, "    incremental %s\n", bool[_ckpt_inc_flag]);
    fprintf(stderr, "    fork %s\n", bool[_ckpt_fork_flag]);
    fprintf(stderr, "    exclude %s\n", bool[_ckpt_exclude_flag]);
    fprintf(stderr, "    verbose %s\n", bool[_ckpt_verbose_flag]);
    fprintf(stderr, "    directory %s\n", _ckpt_dir);
    fprintf(stderr, "    maxfiles %d\n", _ckpt_max_files);
    fprintf(stderr, "    maxtime %d\n", _ckpt_max_time);
    fprintf(stderr, "    mintime %d\n", _ckpt_min_time);
    fprintf(stderr, "    TIME %d\n", time(0));
    fprintf(stderr, "END LIBCKPT DEFAULTS\n");
  }
}

/*
 *  function:        ckpt_setup_()
 *  arguments:       dummy1, dummy2
 *  returns:         void
 *  local variables: none
 *  side effects:    Sets the global variable _ckpt_lang_flag to 1.  This
 *                   lets _ckpt_setup() know it was called from a FORTRAN
 *                   program.  see _ckpt_setup().
 *  calls:           _ckpt_setup()
 *  called from:     fmain.f
 *  misc:            this is called from the FORTRAN main routine,
 *                   just before CKPT_TARGET().  Arguments dummy1 and
 *                   dummy2 are used as placeholders for argc and argv.
 *                   Since this is called from a FORTRAN program, they
 *                   are unused in _ckpt_setup().
 */

void ckpt_setup_(dummy1, dummy2)
int *dummy1;
char **dummy2;
{

  _ckpt_lang_flag = 1;
  _ckpt_setup(dummy1, dummy2);

}

/*
 *  function:        _ckpt_setup()
 *  arguments:       argcp -- a pointer to argc
 *                   argv -- standard usage
 *  returns:         void
 *  local variables: none
 *  side effects:    lots, see the appropriate function calls:
 *                   1 -- sets the flags by reading the .ckptrc files, and
 *                        examining the command line arguments.
 *                   2 -- aborts if checkpoint files exist in the specified dir.
 *                   3 -- initializes global variables
 *                   4 -- "includes" the entire initial address space
 *                   5 -- if incremental checkpointing is enabled:
 *                          a -- sets up the pagelist
 *                          b -- calls signal() to trap SIGSEGV's
 *                          c -- sets the entire address space to READ ONLY,
 *                               then sets the page where the pagelist resides
 *                               back to READ / WRITE
 *                   6 -- sets the alarm if either min_time or max_time != 0
 *  calls:           set_flags()       --  set flags from dot files and argv
 *                   print_defaults()  --  if verbose is on, print useful info 
 *                   check_files()     --  if checkpoint files exist, abort
 *                   set_globals()     --  initialize global variables
 *                   include_bytes()   --  include all memory from start to sbrk
 *                   set_incremental() --  set up incremental checkpointing
 *                   set_alarm()       --  set up timer for asynchronous ckpts
 *  called from:     1 -- ckpt_target_() if FORTRAN
 *                   2 -- main() if C
 */

void _ckpt_setup(argcp, argv)
int *argcp;
char **argv;
{

  set_flags(argcp, argv);

  print_defaults();

  if ( ! _ckpt_flag )
    return;

  set_globals();

#ifndef _pvm
  check_files();
#endif

  include_bytes ( (char *)START_ADDR, (unsigned long)sbrk(0) - START_ADDR );

  if ( _ckpt_inc_flag )
    set_incremental();

  set_alarm();

}

/*
 *  function:        set_flags()
 *  arguments:       argcp    -- a pointer to argc
 *                   argv     -- standard usage
 *  returns:         void
 *  local variables: homepath -- path to user's home directory
 *                   homename -- name of .ckptrc in user's home directory
 *                   flag     -- holds part of line from .ckptrc
 *                   mode     -- holds part of line from .ckptrc
 *  side effects:    sets the global _ckpt flags
 *  calls:           check_for_recovery() -- calls _ckpt_recover() if =recover
 *                                           flag found.  Does NOT return if
 *                                           =recover found!
 *                   set_defaults()       -- sets _ckpt flags to their defaults
 *                   read_dotfile()       -- reads the specified .ckptrc file
 *  called from:     _ckpt_setup()
 *  misc:            1 -- First checks for =recover flag if the language is C.
 *                        (If FORTRAN, this flag is checked for in fmain.f)
 *                        If found, ckpt.0's existance is verified and
 *                        _ckpt_recover() is called.  _ckpt_recover() does NOT
 *                        return, so if called, the remaining steps are NOT
 *                        performed.
 *                   2 -- Global _ckpt flags are assigned their defaults
 *                   3 -- $HOME/.ckptrc is read, possibly changing _ckpt flags.
 *                   4 -- ./.ckptrc is read, possibly changing _ckpt flags.
 *                   5 -- miscellaneous error checking
 *                   6 -- check for =checkpoint flag
 */

static void set_flags(argcp, argv)
int *argcp;
char **argv;
{

  char *homepath;
  char *homename;
  int i;

  check_for_recovery ( argcp, argv );

  set_defaults();

  homepath = getenv("HOME");
  if (homepath != NULL) {
    homename = (char *)malloc(strlen(homepath) + 8);
    strcpy(homename,homepath);
    strcat(homename,"/.ckptrc");
    read_dotfile(homename);
  }

  read_dotfile(".ckptrc");

  /*  miscellaneous error checking */

  if ( _ckpt_max_files == 1 && ! _ckpt_inc_flag )
    _ckpt_coa_flag = FALSE;
  else
    _ckpt_coa_flag = TRUE;

  if ( _ckpt_min_time && _ckpt_max_time && _ckpt_max_time - _ckpt_min_time <= 0 ) {
    fprintf(stderr, "error in .ckptrc: maxtime, if non-zero, ");
    fprintf(stderr, "must be greater than mintime\n");
    _ckpt_flag = FALSE;
    exit(-1);
  }

  /*  If language is FORTRAN, don't check for =checkpoint flag */

  if ( _ckpt_lang_flag == 1 ) return;

/*
 *  Check for "=checkpoint" flag, and if found, turn on _ckpt_flag.
 *  If present, "=checkpoint" must be the last argument.  argc is
 *  decremented to hide its presence from the user.
 */

  for ( i = 1; i < *argcp; i++ )
    if ( ! strcmp ( argv[i], "=checkpoint" ) )
      if ( i == *argcp - 1 ) {
        _ckpt_flag = TRUE;
        (*argcp)--;
      }
      else {
        fprintf(stderr, "=checkpoint flag, if used, must be ");
        fprintf(stderr, "the last argument\n");
        _ckpt_flag = FALSE;
        exit(-1);
      }

}

/*
 *  function:        check_for_recovery()
 *  arguments:       argcp -- pointer to argc
 *                   argv  -- standard usage
 *  returns:         void
 *  local variables: i     -- loop variable
 *                   path  -- path to ckpt.? files
 *                   fd    -- file descriptor for ckpt.0
 *                   fname -- name of ckpt.0 including complete pathname
 *  side effects:    If recovery is specified and all conditions are
 *                   satisfactory, calls _ckpt_recover() which does not
 *                   return, but instead resumes a previous instance of
 *                   the program.
 *  calls:           _ckpt_get_path()
 *                   _ckpt_recover()
 *  called from:     set_flags()
 *  misc:            the logic is as follows:
 *                   if language is "C"
 *                     for all arguments in argv
 *                      if arg = "=recover"
 *                       if argc == 2
 *                        open ckpt.0
 *                         if ckpt.0 does not exist
 *                          print message and exit
 *                         else
 *                          close the file and call recover
 *                       else (argc != 2)
 *                        print message and exit
 */

static void check_for_recovery(argcp, argv)
int *argcp;
char **argv;
{

  int i;
  pname_type path;
  int fd;
  char fname[MAXLEN];

  if ( _ckpt_lang_flag == 0 )

    for ( i = 1; i < *argcp; i++ )
      if ( ! strcmp ( argv[i], "=recover" ) )

#ifdef _pvm
        if ( *argcp == 3 ) {
          _ckpt_filename = strdup(argv[2]);
#else
        if ( *argcp == 2 ) {
          _ckpt_filename = strdup("ckpt");
#endif

          path = _ckpt_get_path();
          sprintf ( fname, "%s/%s.0", path.name, _ckpt_filename );
          fd = OPEN(fname, O_RDONLY);
          if ( fd == -1 ) {
            fprintf(stderr, "=recover flag specified, but %s cannot be opened",
                    fname);
            _ckpt_flag = FALSE;
            exit(-1);
          }
          CLOSE(fd);
          _ckpt_recover(path);
        }

        else { /* *argcp != 2 or 3*/

#ifdef _pvm
          fprintf(stderr,"=recover flag must be followed by the base name\n");
#else
          fprintf(stderr,"=recover flag, if used, must be the only argument\n");
#endif

          _ckpt_flag = FALSE;
          exit(-1);
        }

}

/*
 *  function:        set_defaults()
 *  arguments:       none
 *  returns:         void
 *  local variables: none
 *  side effects:    sets global _ckpt flags to their defaults
 *  calls:           nothing
 *  called from:     set_flags()
 */

static void set_defaults()
{

  _ckpt_flag = TRUE;
  _ckpt_inc_flag = FALSE;
  _ckpt_fork_flag = FALSE;
  _ckpt_enable = TRUE;
  _ckpt_verbose_flag = FALSE;
  _ckpt_exclude_flag = TRUE;
  _ckpt_min_time = 0;
  _ckpt_max_time = 600;
  _ckpt_max_files = 1;

  _ckpt_dir = malloc(strlen(".") + 1);
  strcpy(_ckpt_dir, ".");

}

/*
 *  function:        read_dotfile()
 *  arguments:       fname      -- name of file to open, including pathname
 *  returns:         void
 *  local variables: fd         -- file descriptor for .ckptrc file
 *                   fptr       -- stream ptr for .ckptrc obtained from fdopen
 *                   lnum       -- current line number in .ckptrc
 *                   flag       -- the first word on the current line in .ckptrc
 *                   mode       -- the second word from current line in .ckptrc
 *                   buf        -- the current line in .ckptrc
 *                   flag_array -- an array of values to assign to _ckpt flags
 *                   imode      -- the "mode" from .ckptrc, converted to u long
 *                   iflag      -- the "flag" from .ckptrc, converted to index
 *                   error      -- to hold return value of conver()
 *  side effects:    sets global _ckpt flags based upon .ckptrc file
 *  calls:           init_flag_array()
 *                   convert()
 *  called from:     set_flags()
 *                   _ckpt_get_path()
 *  misc:            flag_array is an array whose elements will contain
 *                   the values to assign to the _ckpt flags as follows:
 *                     flag_array[0] -- _ckpt_flag
 *                     flag_array[1] -- _ckpt_inc_flag
 *                     flag_array[2] -- _ckpt_fork_flag
 *                     flag_array[3] -- _ckpt_enable
 *                     flag_array[4] -- _ckpt_verbose_flag
 *                     flag_array[5] -- _ckpt_exclude_flag
 *                     flag_array[6] -- _ckpt_min_time
 *                     flag_array[7] -- _ckpt_max_time
 *                     flag_array[8] -- _ckpt_max_files
 *                     flag_array[9] -- _ckpt_dir
 *                   convert() is a function which converts the word read
 *                   from the .ckptrc file (i.e. "checkpointing", "fork", etc.)
 *                   into the proper index into flag_array[]; mode is converted
 *                   into imode, where imode is one of TRUE or FALSE for the
 *                   features that can be turned "on" or "off", or a number
 *                   where a number is required, or a pointer to the directory
 *                   name if mode == "directory".
 */

static void read_dotfile(fname)
char *fname;
{

  int fd;
  FILE *fptr;
  int lnum = 0;
  char flag[MAXLEN];
  char mode[MAXLEN];
  char buf[MAXLEN];
  unsigned long flag_array[NUM_FLAGS];
  unsigned long imode;
  int iflag;
  int error;

  fd = OPEN(fname, O_RDONLY);

  if (fd != -1) {

    init_flag_array(flag_array);

    fptr = fdopen(fd,"r");

    while ( fgets ( buf, MAXLEN, fptr ) ) {

      lnum++;
      flag[0] = mode[0] = 0;
      sscanf(buf,"%s %s", flag, mode);

      error = convert(flag, mode, &iflag, &imode);

      if ( error ) {
        fprintf(stderr, "error on line %d of paramenter file %s\n",lnum,fname);
        _ckpt_flag = FALSE;
        exit(-1);
      }

      flag_array[iflag] = imode;

    }
    
    _ckpt_flag = flag_array[0];
    _ckpt_inc_flag = flag_array[1];
    _ckpt_fork_flag = flag_array[2];
    _ckpt_enable = flag_array[3];
    _ckpt_verbose_flag = flag_array[4];
    _ckpt_exclude_flag = flag_array[5];
    _ckpt_min_time = flag_array[6];
    _ckpt_max_time = flag_array[7];
    _ckpt_max_files = flag_array[8];
    _ckpt_dir = (char *)flag_array[9];

    fflush(fptr);
    CLOSE(fd);

  }

}

/*
 *  function:        init_flag_array()
 *  arguments:       flag_array -- ary of long with vals for global _ckpt flags.
 *  returns:         void
 *  local variables: none
 *  side effects:    saves current values of global _ckpt flags in flag_array
 *  calls:           nothing
 *  called from:     read_dotfile()
 */

static void init_flag_array(flag_array)
unsigned long flag_array[];
{
    
  flag_array[0] = _ckpt_flag;
  flag_array[1] = _ckpt_inc_flag;
  flag_array[2] = _ckpt_fork_flag;
  flag_array[3] = _ckpt_enable;
  flag_array[4] = _ckpt_verbose_flag;
  flag_array[5] = _ckpt_exclude_flag;
  flag_array[6] = _ckpt_min_time;
  flag_array[7] = _ckpt_max_time;
  flag_array[8] = _ckpt_max_files;
  flag_array[9] = (unsigned long)_ckpt_dir;

}

/*
 *  function:        convert()
 *  arguments:       flag -- first word from a line in .ckptrc
 *                   mode -- second word from a line in .ckptrc
 *                   iflag -- flag converted to an integer index into flag_array
 *                   imode -- mode converted to a long.  if mode is "off",
 *                            imode is 0; if mode is "on", imode is 1; if mode
 *                            is a value (for maxtime, mintime, etc), imode
 *                            is that value; otherwise (mode is a directory
 *                            name) imode is a pointer to the string holding
 *                            the directory name.
 *  returns:         0 on success, 1 on failure.
 *  local variables: error -- holds the return value
 *  side effects:    if mode is "directory", mallocs space for and copies mode
 *                   to a permanent location.
 *  calls:           isnum()
 *  called from:     read_dotfile()
 */

static int convert(flag, mode, iflag, imode)
char *flag;
char *mode;
int *iflag;
unsigned long *imode;
{

  int error = FALSE;
  char tempname[MAXLEN];

  if ( ! strcmp(flag, "checkpointing") )
    *iflag = 0;
  else if ( ! strcmp(flag, "incremental") )
    *iflag = 1;
  else if ( ! strcmp(flag, "fork") )
    *iflag = 2;
  else if ( ! strcmp(flag, "verbose") )
    *iflag = 4;
  else if ( ! strcmp(flag, "exclude") )
    *iflag = 5;
  else if ( ! strcmp(flag, "mintime") )
    *iflag = 6;
  else if ( ! strcmp(flag, "maxtime") )
    *iflag = 7;
  else if ( ! strcmp(flag, "maxfiles") )
    *iflag = 8;
  else if ( ! strcmp(flag, "directory") )
    *iflag = 9;
  else
    error = TRUE;

  if ( ! error && *iflag <= 5 )
    if ( ! strcmp(mode, "on") )
      *imode = TRUE;
    else if ( ! strcmp(mode, "off") )
      *imode = FALSE;
    else
      error = TRUE;

  if ( ! error && *iflag >= 6 && *iflag <= 8 )
    if ( isnum(mode) )
      *imode = strtol(mode, (char **)NULL, 10);
    else
      error = TRUE;

  if ( ! error && *iflag == 9 ) {
    *imode = (long)malloc ( strlen(mode) + 1 );
    strcpy ( (char *)*imode, mode );
    if (access((char *)*imode, R_OK | W_OK | X_OK))
      error = TRUE;
  }

  return error;

}

/*
 *  function:        _ckpt_get_path()
 *  arguments:       none
 *  returns:         The name of the directory to search for ckpt.? files
 *  local variables: homepath -- the name of the users $HOME directory.
 *                   homename -- the complete name of $HOME/.ckptrc
 *                   path -- the pathname to be returned.
 *  side effects:    sets all global variables found in .ckptrc. Only _ckpt_dir
 *                   is needed.  If called from check_for_recovery() and NO
 *                   recovery takes place, these globals are reset by
 *                   set_flags().  If called from _ckpt_recover(), these values
 *                   are reset during the recovery process.
 *  calls:           read_dotfile()
 *  called from:     check_for_recovery()
 *                   _ckpt_recover()
 *  misc:            1 - Assume path should be the current directory.
 *                   2 - Check $HOME/.ckptrc
 *                   3 - Check ./.ckptrc
 */

pname_type _ckpt_get_path()
{

  char *homepath;
  char *homename;
  pname_type path;

  _ckpt_dir = ".";

  homepath = getenv("HOME");
  if (homepath != NULL) {
    homename = (char *)malloc(strlen(homepath) + 8);
    strcpy(homename,homepath);
    strcat(homename,"/.ckptrc");
    read_dotfile(homename);
  }
  read_dotfile(".ckptrc");

  path.name[0] = 0;
  strcpy ( path.name, _ckpt_dir );

  return path;

}

/*
 *  function:        check_files()
 *  arguments:       none
 *  returns:         void
 *  local variables: dirp  -- pointer to directory
 *                   dp    -- pointer to directory entry
 *                   sflag -- saves global _ckpt_flag
 *  side effects:    checks to see whether any old ckpt.? files exist and
 *                   aborts if any are found.
 *  calls:           nothing
 *  called from:     _ckpt_setup()
 *  misc:            Since opendir() calls open(), but _ckpt_ftable has
 *                   not been initialized yet, the value of _ckpt_flag must
 *                   be saved and _ckpt_flag set to FALSE so that open()
 *                   doesn't try to access the open file table.
 */

static void check_files()
{

  DIR *dirp;
  struct dirent *dp;
  int sflag;

  sflag = _ckpt_flag;
  _ckpt_flag = FALSE;

  if ( (dirp = opendir(_ckpt_dir)) == NULL ) {
    fprintf(stderr, "cannot open directory %s\n", _ckpt_dir);
    _ckpt_flag = FALSE;
    exit(0);
  }

  for (dp = readdir(dirp); dp != NULL; dp = readdir(dirp))
    if ( strstr(dp->d_name, _ckpt_filename) == dp->d_name &&
         strlen(dp->d_name) > strlen(_ckpt_filename) + 1 &&
         dp->d_name[strlen(_ckpt_filename)] == '.' &&
         isnum(dp->d_name + strlen(_ckpt_filename) + 1) ) {
      closedir(dirp);
      fprintf(stderr, "checkpoint files exist in directory %s\n", _ckpt_dir);
#ifdef _pvm
      pvm_exit();
#endif
      exit(-1);
    }

  closedir(dirp);
  _ckpt_flag = sflag;

}

/*
 *  function:        set_globals()
 *  arguments:       none
 *  returns:         void
 *  local variables: none
 *  side effects:    initializes global _ckpt variables.
 *  calls:           make_ml()
 *  called from:     _ckpt_setup()
 */

static void set_globals()
{

  _ckpt_d_old_break = _ckpt_i_old_break = (char *)sbrk(0);
  _ckpt_wait_flag = FALSE;
  _ckpt_filenum = 0;
  _ckpt_num = 0;
  _ckpt_time = time(0);
  _ckpt_mutex = CKPT_ALLOW;

  _ckpt_inc = make_ml();
  _ckpt_sav = make_ml();

#ifdef _pvm
  _ckpt_filename = NULL;          /* the user better call ckpt_name() */
#else
  _ckpt_filename = strdup("ckpt");
#endif

}


/*
 *  function:        set_incremental()
 *  arguments:       none
 *  returns:         void
 *  local variables: stop  -- Value of sbrk(0).
 *                   start -- Beginning address of area to set READ ONLY.
 *                            Does not include first page, since this page
 *                            cannot be mprotect()'d on SparcStations.
 *                   lb    -- lower bound of the page where the pagelist resides
 *                   ub    -- upper bound of the page where the pagelist resides
 *                   test  -- holds the return value of the signal and mprotect
 *                            system calls.  Used for error checking.
 *                   addr  -- used as an argument to _ckpt_markpage() when
 *                            the page holding the pagelist is set to dirty
 *                   i     -- loop variable
 *  side effects:    sets up incremental checkpointing.  An area is malloc()'d
 *                   to specify whether pages are clean or dirty; one bit is
 *                   needed per page.  signal() is called to trap writes to
 *                   clean pages.  The entire address space between start and
 *                   stop is set to READ ONLY.  The pages where the pagelist
 *                   resides is set back to READ / WRITE, since the signal
 *                   hanlder itself writes to these pages.  These pages must
 *                   be marked dirty so they are checkpointed.  When a SEGV is
 *                   caught, the page is marked dirty and the page is set
 *                   to READ / WRITE.
 *  calls:           _ckpt_markpage()
 *  called from:     _ckpt_setup()
 *  misc:            The gross type casts are neccesary because we must
 *                   force the result of division to be a whole number,
 *                   and because pagelist cannot be a long since it's used
 *                   in bitwise operations which require integral operands.
 *                   Finally, the return value of signal() must be cast to
 *                   shut lint up.
 *                   I'd like to make stop, start, addr, etc. of type caddr_t,
 *                   but gcc won't allow them to be used in arithmetic
 *                   operations.  So they're unsigned longs.  sigh.
 */

static void set_incremental()
{
#ifndef __linux
  unsigned long stop;
  unsigned long start;
  unsigned long lb;
  unsigned long ub;
  long test;
  unsigned long addr;
  int i;

  start = START_ADDR + GPAGESIZE;
  stop = (unsigned long)sbrk(0);

  _ckpt_np = (int)((stop - START_ADDR) / GPAGESIZE);

  if (stop % GPAGESIZE) _ckpt_np++;

  _ckpt_pagelist = (unsigned char *)malloc(_ckpt_np);

  if ( _ckpt_pagelist == NULL ) {
    fprintf(stderr, "out of memory\n");
    _ckpt_flag = FALSE;
    exit(-1);
  }

  for ( i = 0; i < _ckpt_np; i++ ) _ckpt_pagelist[i] = 0;

  lb = (unsigned long)((unsigned long)_ckpt_pagelist / GPAGESIZE) * GPAGESIZE;
  ub = ((unsigned long)(((unsigned long)_ckpt_pagelist + _ckpt_np) / GPAGESIZE) + 1) * GPAGESIZE;

  test = (long)signal(SIGSEGV, _ckpt_pgfault_handler);
  if ( test == -1 ) {
    fprintf(stderr, "cannot catch SEGV signals\n");
    _ckpt_flag = FALSE;
    exit(-1);
  }

  test = (long)signal(SIGBUS, _ckpt_pgfault_handler);
  if ( test == -1 ) {
    fprintf(stderr, "cannot catch BUS signals\n");
    _ckpt_flag = FALSE;
    exit(-1);
  }

/*
 *  set memory protection from the end of the first page to the start of
 *  the page containing ckpt_pagelist to PROT_READ.  The first page cannot be
 *  mprotect'd on sparcs.
 */

  if ( stop % GPAGESIZE )
    stop += GPAGESIZE - (stop % GPAGESIZE);

  test = mprotect((caddr_t)start, (size_t)(stop - start), PROT_READ);
  if ( test == -1 ) {
    perror("libckpt");
    fprintf(stderr, "cannot set memory to read only 0x%x, %d.  errno: %d, %d\n",
          start, stop-start, errno, ENOMEM);
    fprintf(stderr, "start: 0x%x, stop: 0x%x\n", start, stop);
    _ckpt_flag = FALSE;
    exit(-1);
  }

/*
 *  The pages spanned by _ckpt_pagelist are written by the page fault
 *  handler, so they must be set to READ / WRITE
 */

  test = mprotect((caddr_t)lb, (size_t)(ub - lb), PROT_READ | PROT_WRITE);
  if ( test == -1 ) {
    fprintf(stderr, "cannot set _ckpt_pagelist to read/write\n");
    _ckpt_flag = FALSE;
    exit(-1);
  }

/*
 *  The first page cannot be mprotect'd on sparcs, so it must always be
 *  considered dirty.  Likewise for the first and last pages spanned by
 *  _ckpt_pagelist if they are partial pages (i.e. _ckpt_pagelist need
 *  not be checkpointed, but if the pages are partial, they may contain
 *  other dirty locations.
 */

  _ckpt_markpage(START_ADDR);

  addr = (unsigned long)_ckpt_pagelist;
  if ( (addr % GPAGESIZE) != 0 ) _ckpt_markpage ( addr );
  if ( (addr + _ckpt_np * sizeof(unsigned char)) % GPAGESIZE != 0 )
    _ckpt_markpage ( addr + _ckpt_np * sizeof(unsigned char));
#endif
}

/*
 *  function:        set_alarm
 *  arguments:       none
 *  returns:         void
 *  local variables: none
 *  side effects:    Calls signal() to catch ALRM signals.
 *                   Calls alarm() to raise SIGALRM for when timer expires
 *  calls:           nothing
 *  called from:     _ckpt_setup()
 */

static void set_alarm()
{

  signal(SIGALRM, _ckpt_alrm_handler);

  if ( _ckpt_min_time ) {
    _ckpt_enable = FALSE;
    alarm(_ckpt_min_time);
  }
  else {
    _ckpt_enable = TRUE;
    if ( _ckpt_max_time ) {
      alarm(_ckpt_max_time);
    }
  }

}

/*
 *  function:        _ckpt_markpage()
 *  arguments:       addr      -- address where SEGV occured.
 *  returns:         void
 *  local variables: pn        -- the page number where SEGV occurred.
 *  side effects:    sets the pn_th element in the global _ckpt_pagelist array
 *  calls:           nothing
 *  called from:     _ckpt_pgfault_handler()
 *                   set_incremental()
 *                   protect_all()
 */

void
_ckpt_markpage(addr)
unsigned long addr;
{

  unsigned long pn;

  pn = (unsigned long)((addr - START_ADDR) / GPAGESIZE);

  _ckpt_pagelist[pn] = 1;

}

/*
 *  function:        isnum()
 *  arguments:       s -- a pointer to a string.
 *  returns:         TRUE if all the characters in s are digits, FALSE otherwise
 *  local variables: none
 *  side effects:    none known
 *  calls:           nothing
 *  called from:     convert()
 */

static int isnum(s)
char *s;
{

  if (!s) return FALSE;

  while ( *s )
    if ( ! isdigit(*s++) )
      return FALSE;

  return TRUE;

}

/*
 *  function:        _ckpt_name()
 *  arguments:       s, a pointer to a NULL terminated string
 *  returns:         0 on success, -1 otherwise and sets EINVAL.  This
 *                   happnes if the argument is a NULL pointer.
 *  local variables: none
 *  side effects:    makes the global _ckpt_filename point to a duplicate of
 *                   the argument
 *  calls:           strdup()
 *  called from:     user's code
 */

int ckpt_name(s)
char *s;
{

    int result;

    if (s != NULL) {
        _ckpt_filename = strdup(s);
        check_files();
        result = 0;
    }

    else {
        result = -1;
        errno = EINVAL;
    }

    return result;

}
